/*********************************************************************
 * AberChat Client 4.20 (C) Giancarlo Castrataro -- iDirt/Compatible *
 *********************************************************************/

#include <stdlib.h>
#include "kernel.h"
#include "client.h"
#include "mobile.h"
#include "sflags.h"
#include "pflags.h"
#include "sendsys.h"
#include "log.h"
#include "mud.h"
#include "parse.h"
#include "bprintf.h"
#include "commands.h"
#include "verbs.h"

Boolean auto_restart = True;
Boolean auth = False;
Boolean aber_output = False;
int aberfd = -1;
char abuffer[WRITELEN];

extern int errno;
extern fd_set sockets_fds;
extern int width;
extern char *WizLevels[];
extern char *MLevels[];
extern char *Flevels[];

/* Chat sequences for all communications lines */

char *chat_chat		= "&=WB[%s&=RB@&=WB%s]&* %s\n";
char *chat_emote	= "&=WB->%s&=RB@&=WB%s %s<-&*\n";
char *god_chat		= "&=WrAGod: %s@%s -->&* %s\n";
char *god_emote		= "&=WrAGod: %s@%s %s&*\n";
char *coder_chat	= "&=WyACode: %s@%s -->&* %s\n";
char *coder_emote	= "&=WyACode: %s@%s %s&*\n";
char *atell_fmt		= "&+R%s&+W@&+c%s&* tells you&+W, &+W'&*%s&+W'&*\n";
char *atell_send_fmt	= "You tell &+R%s&+W@&+c%s&*, &+W'&*%s&+W'&*\n";
char *ainfo_fmt		= "&+b[&+WAberchat info from &+c%s&+W: &*%s&+b]&*\n";
char *areply_fmt	= "You reply, &+W'%s&+W'&*\n";
char *amessage_fmt	= "&+cReply from %s&*: &+W%s\n";

char *player_level(int plr)
{
  if (plev(plr) > LVL_WIZARD)
    return (WizLevels[wlevel(plev(plr))]);
  else {
    if (psex(plr))
      return (FLevels[plev(plr)]);
    else
      return (MLevels[plev(plr)]);
  }
}

int makesock(char *addr, int port)
{
  int s;
  int opt = 1;
  struct hostent *host;
  struct sockaddr_in client;

  if ((host = gethostbyname(addr)) == NULL) {
    return -1;
  }
  client.sin_family = AF_INET;
  client.sin_port = htons(port);
  client.sin_addr = (*(struct in_addr *) host->h_addr_list[0]);

  if ((s = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    return -1;
  }
  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *) &opt, sizeof(opt))
      == -1) {
    close(s);
    return -1;
  }
  if (connect(s, (struct sockaddr *) &client, sizeof(client)) == -1) {
    close(s);
    return (-1);
  }
  if (fcntl(s, F_SETFL, O_NONBLOCK) == -1) {
    close(s);
    return -1;
  }
  return (s);
}


/* Checks to see if a player will hear a line */

Boolean abercheck(int plx, int type)
{
  if (!is_in_game(plx) || !ptstflg(plx, PFL_ABERCHAT) || ststflg(plx, SFL_NOABERCHAT) || !auth)
    return False;
  else {
    switch (type) {
      case CHAT:
        if(ststflg(plx,SFL_NOACHAT)) {
          return False;
        }
        break;
      case ACODE:
        if(!ptstflg(plx, PFL_CODER) || ststflg(plx,SFL_NOACODE)) {
          return False;
        }
        break;
      case AGOD:
        if(plev(plx) < LVL_GOD || ststflg(plx,SFL_NOAGOD)) {
          return False;
        }
        break;
      case INFOMSG:
        if(ststflg(plx, SFL_NOAINFO)) {
          return False;
        }
        break;
      case TELL:
        if(ststflg(plx, SFL_NOATELL)) {
	  return False;
        }
        break;
    }
  }
  return True;
}

/* Pops one argument from a string */

char *get_arg(char *txt)
{
  char *ptr;

  if (txt != NULL && (ptr = strchr(txt, DCHAR))) {
    *ptr = 0;
    return (++ptr);
  } else
    return (NULL);
}

void aautoconcom(void)
{
  if (plev(mynum) < LVL_GOD) {
    bprintf("You aren't powerful enough.\n");
    return;
  }
  if (auto_restart)
    auto_restart = False;
  else
    auto_restart = True;

  bprintf("AberChat Auto Connect is %s.\n", auto_restart ? "On" : "Off");
}

void aboot(Boolean check)
{

  if (check && plev(mynum) < LVL_GOD)
    bprintf("You aren't powerful enough.\n");

  else if (aberfd == -1) {
    if ((aberfd = makesock(SERVER, SERVER_PORT)) == -1) {
      if (check)
	bprintf("Failed.\n");
    } else {
      FD_SET(aberfd, &sockets_fds);
      if (aberfd >= width)
	width = aberfd + 1;
      authenticate();
      if (check)
	bprintf("Connected to AberChat Server.\n");
    }
  } else {
    if (check)
      bprintf("AberChat already running.\n");
  }
}

void ashutdown(Boolean check)
{
  if (check && plev(mynum) < LVL_GOD)
    bprintf("You aren't powerful enough.\n");
  else if (aberfd != -1) {
    ainfo(MYMUD, "Exited Network");
    mudlog("ABERCHAT: Shutdown");
    close(aberfd);
    FD_CLR(aberfd, &sockets_fds);
    aberfd = -1;
  } else
    bprintf("Aberchat not running.\n");
}

void aprintf(char *format,...)
{
  char buffer[WRITELEN];
  Boolean ok = False;
  va_list pvar;

  if (mynum != -1) {
    if (aberfd == -1)
      bprintf("Aberchat not running.\n");
    else if (!auth)
      bprintf("Aberchat not authenticated.\n");
    else if (!ptstflg(mynum, PFL_ABERCHAT))
      bprintf("You aren't authorized to use aberchat.\n");
    else if (ststflg(mynum, SFL_NOABERCHAT))
      bprintf("You don't have AberChat enabled.\n");
    else
      ok = True;
  } else if (aberfd != -1)
    ok = True;

  if (ok) {
    va_start(pvar, format);
    vsprintf(buffer, format, pvar);
    va_end(pvar);
    strcat(abuffer, buffer);
    aber_output = True;
  }
}

void write_aberpacket()
{
  write(aberfd, abuffer, strlen(abuffer) + 1);
  *abuffer = 0;
  aber_output = False;
}

void read_aberpacket()
{
  static char buff[READLEN];
  static char *pos = buff;
  int n_read, data_type;

  n_read = read(aberfd, pos, READLEN - (pos - buff));

  if (n_read <= 0) {
    if (auth) {
      ainfo("AberChat Server", "Aber Server has crashed");
      mudlog("ABERCHAT: Connection closed by Server.");
      ashutdown(False);
    }
  } else if (*(pos + n_read - 1) == 0) {
    data_type = *buff;
    abercmd(*buff, buff + 2);
    *buff = 0;
    pos = buff;
  } else
    pos += n_read;
}


void abercmd(char type, char *packet)
{
  char *arg1, *arg2, *arg3, *arg4;
  int x;

  switch (type) {		/* commands with unlimited args here */
  case MUDLIST:
    amudlist(packet);
    return;
  case RWHO:
    receive_awho(packet);
    return;
  }

  arg1 = packet;
  arg2 = get_arg(arg1);
  arg3 = get_arg(arg2);
  arg4 = get_arg(arg3);
  switch (type) {
  case KICK:
    kick_player(arg1, arg2, arg3);
    return;
  case FILEPAGE:
    afilepage(arg1, arg2, arg3);
    return;
  case AVERSION:
    aversion(arg1, arg2, arg3);
    return;
  case MSG:
    amessage(arg1, arg2, arg3);
    return;
  case TELL:
    atell(arg1, arg2, arg3, arg4);
    return;
  }
  x = mynum;
  mynum = -1;

  switch (type) {
  case AVERREQ:
    aver_reply(arg1, arg2);
    break;
  case WHO:
    send_awho(arg1, arg2);
    break;
  case CHAT:
  case ACODE:
  case AGOD:
    achat(type, arg1, arg2, arg3);
    break;
  case INFOMSG:
    ainfo(arg1, arg2);
    break;
  case AUTH:
    if (*packet == 'Y')
      auth_received(True);
    else
      auth_received(False);
    break;
  }
  mynum = x;
}

void amasslistcom(void)
{
  if (!abercheck(mynum, 0)) {
    bprintf("Can't do that.\n");
    return;
  }
  strcpy((char *) cur_player->work2, "masslist");
  afilepagecom(NULL);
}

void apolicycom(void)
{
  char buff[1024];
  if (!abercheck(mynum, 0)) {
    bprintf("Can't do that.\n");
    return;
  }
  getreinput(buff);
  if (brkword() == -1) {
    strcpy((char *) cur_player->work2, "policy");
  } else {
    sprintf((char *) cur_player->work2, "policy.%s", buff);
  }
  afilepagecom(NULL);
}

void acopyrightcom(void)
{
  if (!abercheck(mynum, 0)) {
    bprintf("Can't do that.\n");
    return;
  }
  strcpy((char *) cur_player->work2, "copyright");
  afilepagecom(NULL);
}

void afilepagecom(char *input)
{
  if (!input) {			/* first run */
    if (!auth || aberfd == -1) {
      bprintf("Aberchat is not running.\n");
      return;
    }
    cur_player->work = 0;
    strcpy(cur_player->cprompt, "");
    push_input_handler(afilepagecom);
  }
  aprintf("%c" D "%s" D "%c" D "%s" D "%s" D "%d", TO_MUD, MYMUD,
	  FILEPAGE, pname(mynum), cur_player->work2, cur_player->work);
  if (!auth || aberfd == -1) {
    strcpy(cur_player->cprompt, build_prompt(mynum));
    bprintf(cur_player->cprompt);
    pop_input_handler();
  }
}

void afilepage(char *player, char *txt, char *more)
{
  int plx;

  if ((plx = fpbn(player)) == -1)
    return;
  else {
    setup_globals(plx);
    bprintf(txt);

    if (*more == '1') {
      cur_player->work++;
      bprintf("[File: %s] ", cur_player->work2);
    } else {
      cur_player->work = 0;
      strcpy(cur_player->cprompt, build_prompt(mynum));
      bprintf(cur_player->cprompt);
      pop_input_handler();
    }
  }
}

/* Prototype: FROM_PLAYER:TO_PLAYER:FROM_MUD:TEXT */

void atell(char *f_plr, char *to_plr, char *f_mud, char *text)
{
  int plx;

  if ((plx = fpbn(to_plr)) != -1 && !pvis(plx)) {
    if (abercheck(plx, TELL)) {
      set_areply(plx, f_plr, f_mud);
      sendf(plx, atell_fmt, f_plr, f_mud, text);
    } else {
      int x = mynum;
      mynum = -1;
      aprintf("%c" D "%s" D "%c" D "%s" D "MUD %s" D "%s isn't on AberChat",
	      TO_MUD, f_mud, MSG, f_plr, MYMUD, to_plr);
      mynum = x;
    }
  } else {
    int x = mynum;
    mynum = -1;
    aprintf("%c" D "%s" D "%c" D "%s" D "MUD %s" D "%s is not online",
	    TO_MUD, f_mud, MSG, f_plr, MYMUD, to_plr);
    mynum = x;
  }
}

void amessage(char *to_plr, char *f_mud, char *text)
{
  int plx;

  if ((plx = fpbn(to_plr)) != -1)
    sendf(plx, amessage_fmt, f_mud, text);
}

void ainfo(char *f_mud, char *text)
{
  int plx;

  for (plx = 0; plx < max_players; plx++)
    if (abercheck(plx, INFOMSG))
      sendf(plx, ainfo_fmt, f_mud, text);
}

void achat(int type, char *f_mud, char *f_plr, char *text)
{
  char *ctl = NULL;
  int plx;

  if (type == CHAT) {
    if (*text == ':')
      ctl = chat_emote;		/* chatline/emote */
    else
      ctl = chat_chat;		/* chatline/chat */
  } else if (type == ACODE) {
    if (*text == ':')
      ctl = coder_emote;	/* coderline/emote */
    else
      ctl = coder_chat;		/* coderline/chat */
  } else if (type == AGOD) {
    if (*text == ':')
      ctl = god_emote;
    else
      ctl = god_chat;
  }
  if (*text == ':')
    text++;

  for (plx = 0; plx < max_players; plx++)
    if (abercheck(plx, type))
      sendf(plx, ctl, f_plr, f_mud, text);
}

/****************************************************************
 * awhocom() sends a WHO request to the specified mud.  The mud *
 * then sends a response packet, in send_awho(), and the packet *
 * is received/displayed in receive_awho()                      *
 ****************************************************************/

void awhocom()
{
  if (EMPTY(txt1))
    bprintf("Which mud?\n");
  else
    aprintf("%c" D "%s" D "%c" D "%s" D "%s",
	    TO_MUD, txt1, WHO, MYMUD, pname(mynum));
}

extern int cmp_player(const void *, const void *);

void send_awho(char *mud, char *player)
{
  char *level, *title;
  int i, list[max_players], len;

  for (i = 0, len = 0; i < max_players; i++)
    if (is_in_game(i) && !pvis(i))
      list[len++] = i;

  qsort(list, len, sizeof(int), cmp_player);

  aprintf("%c" D "%s" D "%c" D "%s" D "%s", TO_MUD, mud, RWHO, player, MYMUD);
  for (i = 0; i < len; i++) {
    level = player_level(list[i]);
    title = make_title(ptitle(list[i]), pname(list[i]));
    aprintf(D "%s" D "%s" D "%s", pname(list[i]), level, title);
  }
}

void receive_awho(char *packet)
{
  char *player, *name, *level, *title, *mud, *next;
  int plx, i;

  player = packet;
  mud = get_arg(player);
  next = get_arg(mud);

  if ((plx = fpbn(player)) == -1)
    return;
  setup_globals(plx);

  if (!next) {
    bprintf("There are no players online at %s.\n%s", mud, cur_player->cprompt);
    return;
  }
  bprintf("\n&+WName          &+CLevel          &+YTitle\n" DASHES);
  for (i = 0; next; i++) {
    name = next;
    level = get_arg(name);
    title = get_arg(level);
    next = get_arg(title);
    bprintf("%-14s%-15s%-45.47s\n", name, level, title);
  }
  bprintf(DASHES "There %s %d player%s on %s.\n%s", (i > 1) ? "are" : "is",
	  i, (i > 1) ? "s" : "", mud, cur_player->cprompt);
}

/***************************************************************************
 * To connect to the network, authenticate() sends a packet which contains *
 * the password, mudname, mudlib, and port.  This allows the server to     *
 * build a mudlist entry for the connecting mud.  Following function is a  *
 * server response, which enables the mud to send data over aberchat       *
 ***************************************************************************/

void authenticate(void)
{
  char passwd[30];
  int oldnum = mynum;

  strcpy(passwd, crypt(PASS, "Ac"));
  auth = True;
  mynum = -1;
  aprintf("%c" D "%s" D "%c" D "%s" D "%s" D "%d" D "%s",
	  TO_MUD, MYMUD, AUTH, MYMUD, MUDLIB, PORT, passwd);
  auth = False;
  mynum = oldnum;
}

void auth_received(Boolean ok)
{
  if (ok) {
    auth = True;
    mudlog("ABERCHAT: Successful connection to server");
    ainfo(MYMUD, "Authenticated/Joined Network");
  } else {
    mudlog("ABERCHAT: Connection failed (Bad password)");
    send_msg(DEST_ALL, MODE_SFLAG | MS(SFL_SEEEXT), LVL_WIZARD, LVL_MAX,
	     NOBODY, NOBODY, "&+C[Aberchat Failed: Bad Password]\n");
    ashutdown(False);
  }
}

/****************************************************************************
 * aversioncom() sends the current version to the aberchat server, or makes *
 * a request to another mud for it's version.  The next function is the     *
 * response, and the 3rd function is where it's displayed.                  *
 ****************************************************************************/

void aversioncom(void)
{
  if (EMPTY(txt1))
    aprintf("%c" D "%s" D "%c" D "%s" D "%s",
	    TO_MUD, MYMUD, AVERSION, pname(mynum), CLIENT_VERSION);
  else {
    bprintf("Version for MUD %s (if no response, client is pre-4.20):\n", txt1);
    aprintf("%c" D "%s" D "%c" D "%s" D "%s",
	    TO_MUD, txt1, AVERREQ, MYMUD, pname(mynum));
  }
}

void aver_reply(char *tomud, char *player)
{
  aprintf("%c" D "%s" D "%c" D "%s" D "%s",
	  TO_MUD, tomud, AVERSION, player, CLIENT_VERSION);
}

void aversion(char *to_plr, char *client_ver, char *server_ver)
{
  int plx;

  if ((plx = fpbn(to_plr)) == -1)
    return;
  sendf(plx, "Server is: %s\n", server_ver);
  sendf(plx, "Client is: %s\n", client_ver);
  sendf(plx, "All code Copyright G. Castrataro (gcastrat@indiana.edu)\n");
}

/* Mudlist request, and response... */

void amudlistcom(void)
{
  aprintf("%c" D "%s" D "%c" D "%s", TO_MUD, MYMUD, MUDLIST, pname(mynum));
}

void amudlist(char *list)
{
  char *toplr, *mudname, *mudlib, *mudaddr, *mudip, *mudport, *next;
  int pl;

  toplr = list;
  next = get_arg(toplr);

  if ((pl = fpbn(toplr)) == -1)
    return;

  setup_globals(pl);

  bprintf("\n&+WMudname         &+CAddress                  &+YIP         "
	  "     &+MPort       &+GCode Base\n" DASHES);

  while (next != NULL) {
    mudname = next;
    mudlib = get_arg(mudname);
    mudaddr = get_arg(mudlib);
    mudip = get_arg(mudaddr);
    mudport = get_arg(mudip);
    next = get_arg(mudport);
    bprintf("%-16.14s%-25.23s%-16.14s%-11s%-11.11s\n",
	    mudname, mudlib, mudaddr, mudip, mudport);
  }
  bprintf("%s%s", DASHES, cur_player->cprompt);
}

void achatcom(int type)
{
  char buff[MAX_COM_LEN];

  if (!abercheck(mynum, type)) {
    bprintf("Sorry, you can't talk on that AberChat line.\n");
    return;
  }
  getreinput(buff);
  if (brkword() == -1)
    bprintf("What would you like to chat?\n");
  else {
    aprintf("%c" D "%c" D "%s" D "%s" D "%s",
	    TO_ALL, type, MYMUD, pname(mynum), buff);
  }
}

void set_areply(int plr, char *target, char *mud)
{
  if (atarget(plr)) {
    FREE(atarget(plr));
  }
  atarget(plr) = COPY(target);
  if (amud(plr)) {
    FREE(amud(plr));
  }
  amud(plr) = COPY(mud);
}

void atellcom(void)
{
  char *mud;

  if (!abercheck(mynum, TELL)) {
    bprintf("Sorry, you can't atell at the moment.\n");
    return;
  }

  if (!(mud = strchr(item1, '@')))
    bprintf("Usage: ATell player@mud [message]\n");
  else {
    *(mud++) = 0;

    set_areply(mynum, item1, mud);

    if (ststflg(mynum, SFL_HEARBACK))
      sendf(mynum, atell_send_fmt, item1, mud, txt2);
    aprintf("%c" D "%s" D "%c" D "%s" D "%s" D "%s" D "%s\0",
	    TO_MUD, mud, TELL, pname(mynum), item1, MYMUD, txt2);
  }
}

void areplycom(void)
{
  if (EMPTY(txt1)) {
    bprintf("Reply what? You need to type something after it.\n");
    return;
  } else if (!amud(mynum) || !atarget(mynum)) {
    bprintf("You have no-one to reply to.\n");
    return;
  } else if (!abercheck(mynum, TELL)) {
    bprintf("Sorry, you can't areply at the moment.\n");
    return;
  }
  if (ststflg(mynum, SFL_HEARBACK))
    sendf(mynum, areply_fmt, txt1);
  aprintf("%c" D "%s" D "%c" D "%s" D "%s" D "%s" D "%s\0", TO_MUD,
	  amud(mynum), TELL, pname(mynum), atarget(mynum), MYMUD, txt1);
}

void ainfocom(void)
{
  char buff[MAX_COM_LEN];

  if (!abercheck(mynum, INFOMSG)) {
    bprintf("Sorry, you can't send AINFO messages.\n");
    return;
  }
  getreinput(buff);
  if (brkword() == -1)
    bprintf("What should the message be?\n");
  else
    aprintf("%c" D "%c" D "%s" D "%s", TO_ALL, INFOMSG, MYMUD, buff);
}

void kick_player(char *from, char *to, char *frommud)
{
  int plr;

  if ((plr = fpbn(to)) == -1)
    aprintf("%c" D "%s" D "%c" D "%s" D "MUD %s" D "%s is not online",
	    TO_MUD, frommud, MSG, from, MYMUD, to);
  else if (plev(plr) > LVL_GOD)
    aprintf("%c" D "%s" D "%c" D "%s" D "MUD %s" D "%s is too powerful",
	    TO_MUD, frommud, MSG, from, MYMUD, to);
  else {
    sendf(plr, "You have been kicked off aberchat by %s@%s\n", from, frommud);
    pclrflg(plr, PFL_ABERCHAT);
    pclrmsk(plr, PFL_ABERCHAT);
    mudlog("%s has been akicked by %s@%s.",pname(plr),from,frommud);
  }
}

void akickcom(void)
{
  char *mud;

  if (plev(mynum) < LVL_GOD)
    bprintf("You aren't powerful enough.\n");
  else if (!(mud = strchr(item1, '@')))
    bprintf("Usage: AKick player@mud\n");
  else {
    *(mud++) = 0;
    aprintf("%c" D "%s" D "%c" D "%s" D "%s" D "%s",
	    TO_MUD, mud, KICK, pname(mynum), item1, MYMUD);
  }
}

void noaberchatcom(void)
{
  if (!ptstflg(mynum, PFL_ABERCHAT))
    bprintf("You can't use aberchat.\n");
  else
    togglecom(SFL_NOABERCHAT, "Aberchat Disabled.", "Aberchat Enabled.");
  return;
}

void noachatcom(void)
{
  if (!ptstflg(mynum, PFL_ABERCHAT)) {
    bprintf("You can't use the achat line.\n");
  } else {
    togglecom(SFL_NOACHAT, "AChat line disabled.","AChat line enabled.");
  }
  return;
}

void noacodecom(void)
{
  if (!ptstflg(mynum, PFL_ABERCHAT) || !ptstflg(mynum, PFL_CODER)) {
    bprintf("You can't use the acode line.\n");
  } else {
    togglecom(SFL_NOACODE, "ACode line disabled.","ACode line enabled.");
  }
  return;
}

void noagodcom(void)
{
  if (!ptstflg(mynum, PFL_ABERCHAT) || plev(mynum) < LVL_GOD) {
    bprintf("You can't use the agod line.\n");
  } else {
    togglecom(SFL_NOAGOD, "AGod line disabled.","AGod line enabled.");
  }
  return;
}

void noainfocom(void)
{
  if (!ptstflg(mynum, PFL_ABERCHAT)) {
    bprintf("You can't see the ainfo line.\n");
  } else {
    togglecom(SFL_NOAINFO, "AInfo line disabled.","AInfo line enabled.");
  }
  return;
}

void noatellcom(void)
{
  if (!ptstflg(mynum, PFL_ABERCHAT)) {
    bprintf("You can't see atells.\n");
  } else {
    togglecom(SFL_NOATELL, "You won't receive atells.","You will receive atells.");
  }
  return;
}

Boolean aberchat_parse(int vb)
{
  switch(vb) {
  case VERB_NOACHAT: noachatcom(); break;
  case VERB_NOACODE: noacodecom(); break;
  case VERB_NOAGOD: noagodcom(); break;
  case VERB_NOAINFO: noainfocom(); break;
  case VERB_NOATELL: noatellcom(); break;
  case VERB_ABOOT: aboot (True); break;
  case VERB_ASHUTDOWN: ashutdown (True); break;
  case VERB_AWHO: awhocom (); break;
  case VERB_NOABERCHAT: noaberchatcom (); break;
  case VERB_AKICK: akickcom (); break;
  case VERB_AAUTOCON: aautoconcom (); break;
  case VERB_AMUDLIST: amudlistcom (); break;
  case VERB_AVERSION: aversioncom (); break;
  case VERB_AMASSLIST: amasslistcom (); break;
  case VERB_APOLICY: apolicycom (); break;
  case VERB_ACOPYRIGHT: acopyrightcom (); break;
  case VERB_ATELL: atellcom (); break;
  case VERB_AINFO: ainfocom (); break;
  case VERB_ACHAT: achatcom (CHAT); break;
  case VERB_ACODE: achatcom(ACODE); break;
  case VERB_AGOD: achatcom(AGOD); break;
  case VERB_AREPLY: areplycom(); break;
  default: return False;
  }
  return True;
}
